@@ -19,7 +19,7 @@
 #'
 #' @param packages A character vector, indicating package names which should be
 #'   installed or removed. Use  \verb{<package>==<version>} to request the installation
-#'   of a specific version of a package. A `NULL` value for [conda_remove()] 
+#'   of a specific version of a package. A `NULL` value for [conda_remove()]
 #'   will be interpretted to `"--all"`, removing the entire environment.
 #'
 #' @param environment The path to an environment definition, generated via
@@ -104,11 +104,18 @@ conda_list <- function(conda = "auto") {
   conda <- conda_binary(conda)
   local_conda_paths(conda)
 
+  if (startsWith(basename(conda), "micromamba")) {
+    conda_envs <- system2(conda, c("env", "list", "--json"),
+                          stdout = TRUE, stderr = FALSE)
+
+  } else {
   # list envs -- discard stderr as Anaconda may emit warnings that can
   # otherwise be ignored; see e.g. https://github.com/rstudio/reticulate/issues/474
-  conda_envs <- suppressWarnings(
-    system2(conda, args = c("info", "--json"), stdout = TRUE, stderr = FALSE)
-  )
+    conda_envs <- suppressWarnings(
+      system2(conda, args = c("info", "--json"),
+              stdout = TRUE, stderr = FALSE)
+    )
+  }
 
   # check for error
   status <- attr(conda_envs, "status") %||% 0L
@@ -190,6 +197,8 @@ conda_create <- function(envname = NULL,
                          conda = "auto",
                          python_version = miniconda_python_version())
 {
+  check_forbidden_install("Conda Environments")
+
   # resolve conda binary
   conda <- conda_binary(conda)
   local_conda_paths(conda)
@@ -241,6 +250,8 @@ conda_create <- function(envname = NULL,
 
 conda_create_env <- function(envname, environment, conda) {
 
+  check_forbidden_install("Conda Environments")
+
   if (!is.null(envname))
     envname <- condaenv_resolve(envname)
 
@@ -508,7 +519,7 @@ conda_binary <- function(conda = "auto") {
   }
 
   conda <- normalizePath(conda, winslash = "/", mustWork = FALSE)
-  if (!grepl("^(conda|mamba)", basename(conda)))
+  if (!grepl("^(conda|mamba|micromamba)", basename(conda)))
     warning("Supplied path is not a conda binary: ", sQuote(conda))
 
   # if the user has requested a conda binary in the 'condabin' folder,
@@ -630,6 +641,70 @@ conda_python <- function(envname = NULL,
   path.expand(python)
 }
 
+#' @returns
+#'   `conda_search()` returns an \R `data.frame` describing packages that
+#'   matched against `matchspec`. The data frame will usually include
+#'   fields `name` giving the package name, `version` giving the package
+#'   version, `build` giving the package build, and `channel` giving the
+#'   channel the package is hosted on.
+#'
+#' @param matchspec A conda MatchSpec query string.
+#'
+#' @rdname conda-tools
+#' @export
+conda_search <-
+  function(matchspec,
+           forge = TRUE,
+           channel = character(),
+           conda = "auto",
+           ...) {
+
+  conda <- conda_binary(conda)
+  local_conda_paths(conda)
+
+  args <- c("search", matchspec)
+
+  # add user-requested channels
+  channels <- if (length(channel))
+    channel
+  else if (forge)
+    "conda-forge"
+
+  for (ch in channels)
+    args <- c(args, "-c", ch)
+
+  args = c(args, "--json")
+  output <- system2(conda, shQuote(args), stdout = TRUE)
+  status <- attr(output, "status") %||% 0L
+  # check for errors
+  if (status != 0L) {
+    fmt <- "error searching for packages [error code %i]"
+    stopf(fmt, status)
+  }
+
+  parsed <- jsonlite::fromJSON(output)
+
+  all_colnames <- unique(unlist(lapply(parsed, names)))
+  df <- do.call(rbind, lapply(parsed, function(x) {
+    if(any(missing_cols <- setdiff(all_colnames, names(x))))
+      x[missing_cols] <- NA
+    x
+  }))
+  rownames(df) <- NULL
+
+  # reverse row order, so most recent versions are (likely) first. In an ideal
+  # world we'd use `order(numeric_version(df$version))`, but we can't depend on
+  # the version string being parseable.
+  if(nrow(df))
+    df <- df[rev(seq_len(nrow(df))), ]
+
+  # reorder columns
+  to_front <- intersect(c("name", "version", "build", "channel"), names(df))
+  df <- df[c(unique(to_front, names(df)))]
+
+  df
+}
+
 
 find_conda <- function() {
 
@@ -725,6 +800,8 @@ condaenv_path <- function(envname = NULL) {
 
 }
 
+#' @export
+#' @rdname conda-tools
 condaenv_exists <- function(envname = NULL, conda = "auto") {
 
   # check that conda is installed
@@ -884,7 +961,6 @@ conda_run2_nix <-
            intern = FALSE, echo = !intern) {
   conda <- normalizePath(conda_binary(conda))
   local_conda_paths(conda)
-  activate <- normalizePath(file.path(dirname(conda), "activate"))
 
   if (!identical(envname, "base")) {
     envname <- condaenv_resolve(envname)
@@ -895,6 +971,26 @@ conda_run2_nix <-
   fi <- tempfile(fileext = ".sh")
   on.exit(unlink(fi))
 
+  stdout <- if (identical(intern, FALSE)) "" else intern
+
+  if (startsWith(basename(conda), "micromamba")) {
+
+    envflag <- if(grepl("[/\\]", envname)) "-p" else "-n"
+    if(echo)
+      system2 <- system2t
+    result <- system2(conda, c('run', envflag, envname, cmd_line),
+                      stdout = stdout)
+
+    error_status <- attr(result, "status")
+    if (!is.null(error_status))
+      stop("Error ", error_status, " occurred while running conda command")
+
+    return(result)
+
+  }
+
+
+  activate <- normalizePath(file.path(dirname(conda), "activate"))
   commands <- c(
     paste(".", activate),
     if (!identical(envname, "base"))
@@ -910,8 +1006,8 @@ conda_run2_nix <-
       commands))
 
   writeLines(commands, fi)
-  system2(Sys.which("bash"), fi,
-          stdout = if (identical(intern, FALSE)) "" else intern)
+  system2(Sys.which("bash"), fi, stdout = stdout)
+
 }
 
 
