@@ -4,26 +4,36 @@
 #' R interface to Python modules, classes, and functions. When calling into
 #' Python R data types are automatically converted to their equivalent Python
 #' types. When values are returned from Python to R they are converted back to R
-#' types. The reticulate package is compatible with all versions of Python >= 2.7.
+#' types. The reticulate package is compatible with all versions of Python >= 3.6.
 #' Integration with NumPy requires NumPy version 1.6 or higher.
 #'
-#' @docType package
 #' @name reticulate
+#' @aliases reticulate-package
 #' @keywords internal
 #' @useDynLib reticulate, .registration = TRUE
 #' @importFrom Rcpp evalCpp
-NULL
+"_PACKAGE"
 
 # package level mutable global state
 .globals <- new.env(parent = emptyenv())
 .globals$required_python_version <- NULL
 .globals$use_python_versions <- c()
 .globals$py_config <- NULL
-.globals$delay_load_module <- NULL
-.globals$delay_load_environment <- NULL
-.globals$delay_load_priority <- 0
+.globals$delay_load_imports <- data.frame(module = character(),
+                                          priority = integer(),
+                                          environment = character(),
+                                          stringsAsFactors = FALSE)
 .globals$suppress_warnings_handlers <- list()
-.globals$class_filters <- list()
+.globals$class_filters <- list(
+
+  function(classes) {
+    if ("python.builtin.BaseException" %in% classes) {
+      classes <- unique(c(classes, "error", "condition"))
+    }
+    classes
+  }
+
+)
 .globals$py_repl_active <- FALSE
 
 is_python_initialized <- function() {
@@ -37,26 +47,27 @@ ensure_python_initialized <- function(required_module = NULL) {
   if (is_python_initialized())
     return()
 
-  # give delay load modules priority
-  use_environment <- NULL
-  if (!is.null(.globals$delay_load_module)) {
-    required_module <- .globals$delay_load_module
-    use_environment <- .globals$delay_load_environment
-    .globals$delay_load_module <- NULL # one shot
-    .globals$delay_load_environment <- NULL
-    .globals$delay_load_priority <- 0
-  }
-
   # notify front-end (if any) that Python is about to be initialized
   callback <- getOption("reticulate.python.beforeInitialized")
   if (is.function(callback))
     callback()
 
+  # make sure this module is used for an environment name.
+  if(!is.null(required_module))
+    register_delay_load_import(required_module)
+
   # perform initialization
-  .globals$py_config <- initialize_python(required_module, use_environment)
+  .globals$py_config <- initialize_python()
+
+  # clear the global list of delay_load requests
+  .globals$delay_load_imports <- NULL
 
   # remap output streams to R output handlers
   remap_output_streams()
+  set_knitr_python_stdout_hook()
+
+  if (is_windows() && identical(.Platform$GUI, "RStudio"))
+    import("rpytools.subprocess")$patch_subprocess_Popen()
 
   # generate 'R' helper object
   py_inject_r()
@@ -101,13 +112,13 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
 
   # provide hint to install Miniconda if no Python is found
   python_not_found <- function(msg) {
-    hint <- "Use reticulate::install_miniconda() if you'd like to install a Miniconda Python environment."
+    hint <- "Please create a default virtual environment with `reticulate::virtualenv_create('r-reticulate')`."
     stop(paste(msg, hint, sep = "\n"), call. = FALSE)
   }
 
   # resolve top level module for search
   if (!is.null(required_module))
-    required_module <- strsplit(required_module, ".", fixed = TRUE)[[1]][[1]]
+    required_module <- strsplit(required_module, ".", fixed = TRUE)[[1L]][[1L]]
 
   # find configuration
   config <- local({
@@ -237,10 +248,21 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
 
   }
 
-  if (is_windows()) {
+  local({
     # patch sys.executable to point to python.exe, not Rterm.exe or rsession-utf8.exe, #1258
-    py_run_string_impl("import sys; sys.executable = sys.argv[0]", local = TRUE)
-  }
+    patch <- sprintf("import sys; sys.executable  = r'''%s'''",
+                     config$executable)
+    py_run_string_impl(patch, local = TRUE)
+  })
+
+  if (nzchar(config$base_executable)) local({
+    # just like sys.executable, patch to point to python.exe, not Rterm.exe
+    # need to patch for multiprocessing to work on windows, perhaps other things too.
+    # in venvs, _base_executable should point to the venv starter, #1430
+    patch <- sprintf("import sys; sys._base_executable = r'''%s'''",
+                     config$base_executable)
+    py_run_string_impl(patch, local = TRUE)
+  })
 
   # ensure modules can be imported from the current working directory
   py_run_string_impl("import sys; sys.path.insert(0, '')", local = TRUE)
@@ -287,6 +309,8 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
   config
 }
 
+# unused presently, formerly called from initialize_python()
+# https://github.com/rstudio/reticulate/commit/e8c82a1f95eb97c4e5fc27b6550a4498827438e0#r122213856
 check_forbidden_initialization <- function() {
 
   if (is_python_initialized())
