@@ -20,22 +20,32 @@ pip_version <- function(python) {
 }
 
 pip_install <- function(python,
-                        packages,
+                        packages = NULL,
                         pip_options = character(),
                         ignore_installed = FALSE,
                         conda = "auto",
-                        envname = NULL)
+                        envname = NULL,
+                        requirements = NULL,
+                        no_user = TRUE)
 {
   # construct command line arguments
-  args <- c("-m", "pip", "install", "--upgrade")
+  args <- c("-m", "pip", "install", "--upgrade",
+            if (no_user) "--no-user")
   if (ignore_installed)
     args <- c(args, "--ignore-installed")
   args <- c(args, pip_options)
 
+
+  if (!is.null(packages)) {
   # quote in case of version constraints like 'Pillow<8.3'
   # but don't double quote if already quoted
-  packages <- shQuote(gsub("[\"']", "", packages))
-  args <- c(args, packages)
+    packages <- maybe_shQuote(gsub("[\"']", "", packages))
+    args <- c(args, packages)
+  }
+
+  if (!is.null(requirements)) {
+    args <- c(args, "-r", maybe_shQuote(requirements))
+  }
 
   # figure out if we should go though conda_run()
   if (conda == "auto") {
@@ -45,7 +55,7 @@ pip_install <- function(python,
   }
 
   result <- if (is.null(conda) || identical(conda, FALSE))
-    system2(python, args)
+    system2t(python, args)
   else
     conda_run2(python, args, conda = conda, envname = envname)
 
@@ -63,7 +73,7 @@ pip_uninstall <- function(python, packages) {
 
   # run it
   args <- c("-m", "pip", "uninstall", "--yes", packages)
-  result <- system2(python, args)
+  result <- system2t(python, args)
   if (result != 0L) {
     pkglist <- paste(shQuote(packages), collapse = ", ")
     msg <- paste("Error removing package(s):", pkglist)
