@@ -149,8 +149,13 @@ repl_python <- function(
   handle_error <- function(output) {
     failed <- inherits(output, "error")
     if (failed) {
-      error <- py_last_error()
-      message(paste(error$type, error$value, sep = ": "))
+      error_message <- py_last_error()$message
+
+      if (identical(.Platform$GUI, "RStudio") &&
+          requireNamespace("cli", quietly = TRUE))
+        error_message <- make_filepaths_clickable(error_message)
+
+      message(error_message, appendLF = !endsWith(error_message, "\n"))
     }
     failed
   }
@@ -164,15 +169,7 @@ repl_python <- function(
   repl <- function() {
 
     # flush stdout, stderr on each REPL iteration
-    on.exit({
-
-      if (!is.null(sys$stdout) && !is.null(sys$stdout$flush))
-        sys$stdout$flush()
-
-      if (!is.null(sys$stderr) && !is.null(sys$stderr$flush))
-        sys$stderr$flush()
-
-    }, add = TRUE)
+    on.exit(py_flush_output(), add = TRUE)
 
     # read input (either from user or from code)
     prompt <- if (buffer$empty()) ">>> " else "... "
@@ -250,7 +247,7 @@ repl_python <- function(
       }
 
       # similar handling for help requests postfixed with '?'
-      if (grepl("[?]\\s*$", trimmed)) {
+      if (grepl("(^[\\#].*)[?]\\s*$", trimmed, perl = TRUE)) {
         replaced <- sub("[?]\\s*$", "", trimmed)
         code <- sprintf("help(\"%s\")", replaced)
         py_run_string(code)
@@ -572,3 +569,46 @@ invoke_magic <- function(command) {
 }
 
 
+
+#' IPython console
+#'
+#' Launch IPython console app.
+#'
+#' See https://ipython.readthedocs.io/ for features.
+#'
+#' @keywords internal
+ipython <- function() {
+  ensure_python_initialized("IPython")
+
+  # set flag for frontend
+  .globals$py_repl_active <- TRUE
+  on.exit({
+    .globals$py_repl_active <- FALSE
+  }, add = TRUE)
+
+  # don't pollute R history w/ python commands
+  # (IPython keeps track of it's own history)
+  use_history <-
+    !"--vanilla" %in% commandArgs() &&
+    !"--no-save" %in% commandArgs() &&
+    !is.null(getwd()) &&
+    tryCatch(
+      { utils::savehistory(tempfile()); TRUE },
+      error = function(e) FALSE
+    )
+
+  if (use_history) {
+    # if we have history, save and then restore the current
+    # R history
+    utils::savehistory()
+    on.exit(utils::loadhistory(), add = TRUE)
+  }
+
+  # Not implemented,
+  #  - custom startup banner
+  # RStudio IDE support for:
+  #  - image display
+  #  - composition of multi-line code blocks
+
+  import("rpytools.ipython")$start_ipython()
+}
