@@ -1,6 +1,7 @@
 context("envs")
 
 test_that("conda utility functions work as expected", {
+  # TODO: reenable these tests
   skip_if_no_test_environments()
 
   binary <- conda_binary()
@@ -17,12 +18,23 @@ test_that("conda utility functions work as expected", {
   conda_remove('reticulate-testthat')
   expect_false('reticulate-testthat' %in% conda_list()$name)
 
+  conda_create('reticulate-testthat', forge = TRUE)
+  expect_true(all(grepl("conda-forge", conda_list_packages("reticulate-testthat")$channel)))
+  conda_remove('reticulate-testthat')
+
+  conda_create('reticulate-testthat', channel = c("anaconda"))
+  expect_true(all(grepl("anaconda", conda_list_packages("reticulate-testthat")$channel)))
+  conda_remove('reticulate-testthat')
+
 })
 
 test_that("virtualenv utility functions work as expected", {
   skip_if_no_test_environments()
 
-  virtualenv_remove('reticulate-testthat', confirm = FALSE)
+  expect_error(
+    virtualenv_remove('reticulate-testthat', confirm = FALSE),
+    'Virtual environment \'reticulate-testthat\' does not exist.'
+  )
 
   virtualenv_create('reticulate-testthat')
   virtualenv_remove('reticulate-testthat', confirm = FALSE)
